from typing import Optional
from pathlib import Path

import re
import yaml
import pandas as pd

from pydantic import BaseModel, ValidationError, validator

# use this over Literal to make custom error containing more info
ALLOWED_TYPES = ["float", "int", "categorical"]
ROOT_DIR = Path(__file__).parent.parent


def parse_from_yaml(path_to_yaml):
    """
    ability to parse config.yaml
    """
    print(ROOT_DIR)
    with open(path_to_yaml) as f:
        config = yaml.safe_load(f)

    return config


def validate_color_string(color_string):
    pattern = r"^\d{1,3},\s?\d{1,3},\s?\d{1,3}$"
    return re.match(pattern, color_string) is not None


class FeatureSchema(BaseModel):
    name: str
    type: Optional[str]
    n_bins: Optional[int] = 10
    bins: Optional[list[int]] = None

    # needs to be hashable: https://stackoverflow.com/questions/63721614/unhashable-type-in-fastapi-request
    class Config:
        frozen = True

    @validator("type")
    def validate_type(cls, v, values):
        if v not in ALLOWED_TYPES:
            raise ValueError(
                f'Type for {values["name"]} must be in {ALLOWED_TYPES}, you passed type: {v}'
            )
        return v

    @validator("bins")
    def validate_bins_and_nbins(cls, bins, values):
        if len(bins) != values["n_bins"]:
            values["n_bins"] = len(bins)
            print(f'setting n_bins for {values["name"]} to {len(bins)} based on bins')
        return bins


class Settings(BaseModel):
    file_path: Path
    images_output_path: Path
    powerpoint_output_path: Path
    features: list[FeatureSchema]
    target: str
    primary_color: str
    secondary_color: str
    tertiary_color: str
    grey_tint_color: str

    @validator("file_path", "images_output_path", pre=True)
    def path_must_exist(cls, v):
        """
        validates path specified in config. the path/file must exist before the code can execute the ensure valid reading and writing locations exist
        """

        if not isinstance(v, Path):
            v = Path(v)

        if not v.exists():
            raise ValueError(
                f"Directory: {v} does not exist \n Please ensure the config.yaml contains a valid directory"
            )
        return v

    @validator("powerpoint_output_path")
    def parent_dir_must_exist(cls, v):
        """
        validates path specified in config. the path/file must exist before the code can execute the ensure valid reading and writing locations exist
        """

        if not isinstance(v, Path):
            v = Path(v)

        if not v.parent.exists():
            raise ValueError(
                f"Directory: {v.parent} does not exist \n Please ensure the config.yaml contains a valid directory"
            )
        return v

    @validator(
        "primary_color",
        "secondary_color",
        "tertiary_color",
        "grey_tint_color",
        pre=True,
    )
    def test_color_pattern(cls, v):
        """
        matches the string pattern provided against a format of 3 digits, 3 digits, 3 digits
        """

        if not validate_color_string(v):
            raise ValueError(f"Colors must be of format DDD, DDD, DDD (now {v})")
        return v

    @validator("features")
    def check_columns_exist(cls, v, values):
        """
        validates that each of the columns provided in the config exists in the dataframe
        """

        # can only use file path if test was succesfull
        if "file_path" not in values.keys():
            raise ValueError("cannot test features because file_path invalid")
        sample_df = pd.read_csv(values["file_path"], nrows=10)

        for feature in v:
            if feature.name not in sample_df.columns:
                raise ValueError(
                    f"Column: {feature} does not exist is dataframe \n Please ensure the config.yaml contains only valid columns"
                )

        return v

    @validator("target")
    def check_target_exists(cls, v, values):
        """
        validates that the target exists in the dataframe
        """

        # can only use file path if test was succesfull
        if "file_path" not in values.keys():
            raise ValueError("cannot test features because file_path invalid")
        sample_df = pd.read_csv(values["file_path"], nrows=10)

        if v not in sample_df.columns:
            raise ValueError(
                f"Column: {v} does not exist is dataframe \n Please ensure the config.yaml contains only valid columns"
            )
        return v


# create a config object containing all the validated settings
config = parse_from_yaml(str(ROOT_DIR / "config.yaml"))
settings = Settings(**config)

# for debugging purposes
if __name__ == "__main__":
    config = parse_from_yaml(str(ROOT_DIR / "config.yaml"))
    print(config)
    settings = Settings(**config)
    print(settings)



import logging
import pathlib
import argparse
from datetime import datetime

import pandas as pd

from plottah.plot_builder import build_univariate_plots, build_powerpoint
from plottah.config import settings
from plottah.colors import PlotColors

parser = argparse.ArgumentParser(description="Run the univariate analyses workflow")
logging.basicConfig(level=logging.WARN)


def main():
    # parse command line arguments
    parser.add_argument(
        "-p",
        "--build_powerpoint",
        nargs="?",
        const=1,
        type=int,
        default=0,
        choices=[0, 1],
        help="Specify whether you would like to automatically add figures to a powerpoint",
    )

    args = parser.parse_args()
    build_pp = args.build_powerpoint

    # set color palette to use
    color_palette = PlotColors(
        primary_color=settings.primary_color,
        secondary_color=settings.secondary_color,
        tertiary_color=settings.tertiary_color,
        grey_tint_color=settings.grey_tint_color,
    )

    # parse feature names from settings
    features = [feature_schema.name for feature_schema in settings.features]

    # create mapping from feature name to binning
    bins = {
        feature_schema.name: feature_schema.bins
        if feature_schema.bins is not None
        else None
        for feature_schema in settings.features
    }

    # create mapping from feature name to number of bins
    n_bins = {
        feature_schema.name: feature_schema.n_bins
        for feature_schema in settings.features
    }

    # create mapping from feature name to feature type
    feature_types = {
        feature_schema.name: feature_schema.type
        if feature_schema.type is not None
        else "float"
        for feature_schema in settings.features
    }

    # build all the univariate plots
    figs, fig_locs = build_univariate_plots(
        df=pd.read_csv(settings.file_path),
        features=features,
        target=settings.target,
        feature_types=feature_types,
        save_directory=settings.images_output_path,
        colors=color_palette,
        bins=bins,
        n_bins=n_bins,
    )

    if build_pp:
        now = datetime.now()
        current_time = now.strftime("%d_%m_%Y_%H_%M_%S")
        save_path = pathlib.Path(
            f"./data/powerpoints/univariate_analyses_{current_time}.pptx"
        ).resolve()
        print(f"Building output powerpoint in location {save_path}")
        build_powerpoint(
            fig_locs=fig_locs,
            feature_names=features,
            save_path=save_path,
        )


if __name__ == "__main__":
    main()



from typing import Dict

import pathlib

from pptx import Presentation
from pptx.util import Inches, Pt

from plottah.plot_handler import PlotHandler
from plottah.colors import PlotColors
from plottah.plot_builder.specific_builders import PLOT_BUILDERS_DICT


def build_univariate_plot(
    df,
    feature_col: str,
    target: str,
    feature_type: str = "float",
    colors: PlotColors = PlotColors(),
    show_plot: bool = False,
    hoverinfo="all",
    n_bins: int = 10,
    bins: list = None,
    specs: list = None,
):
    """
    buils standard univariate plot from days 'ye

    Returns
    """

    # if feature type not-categoriocal, assume numerical plot type
    feature_type = feature_type if feature_type == "categorical" else "numerical"

    # get appropriate plot builder
    plot_builder = PLOT_BUILDERS_DICT[feature_type]

    # build plot
    plot = plot_builder(
        df,
        feature_col=feature_col,
        target=target,
        feature_type=feature_type,
        colors=colors,
        show_plot=show_plot,
        hoverinfo=hoverinfo,
        n_bins=n_bins,
        bins=bins,
        specs=specs,
    )

    # show the plot if show_plot set to true
    if show_plot:
        plot.show()

    return plot


def build_univariate_plots(
    df,
    features: list,
    target: str,
    feature_types: dict,
    n_bins: dict = None,
    bins: dict = None,
    save_directory: pathlib.Path() = None,
    colors: PlotColors = PlotColors(),
    show_plot: bool = False,
    hoverinfo="none",
) -> Dict[str, PlotHandler]:
    """
    function that generates standard univariate plots

    Args:
        df (pd.DataFrame): dataframe containing columns to analyze,
        features (list): list of columns,
        save_directory (pathlib.Path()): where to store figures:

    Returns:
        Dict: map from feature name to figure

    """
    # if only a single feature passed wrap in a list
    if isinstance(features, str):
        features = [features]

    # create mapping from bins to none if not passed as argument
    if bins is None:
        bins = {feature: None for feature in features}

    # create mapping from features to 10 if nbins not passed as argument
    if n_bins is None:
        n_bins = {feature: 10 for feature in features}
    else:
        for feature in features:
            # user does not have to provide complete mapping
            if feature not in n_bins.keys():
                n_bins[feature] = 10

    # create mapping from features to float if type not provided
    if feature_types is None:
        feature_types = {feature: "float" for feature in features}

    # Run loop
    figs = {}
    save_locs = []
    for i, feature in enumerate(features):
        print(f"[{i+1}/{len(features)}] Starting univariate analysis for: {feature}")
        if feature not in df.columns:
            raise ValueError(f"{feature} not in columns of dataframe")

        fig = build_univariate_plot(
            df=df,
            feature_col=feature,
            target=target,
            feature_type=feature_types[feature],
            colors=colors,
            show_plot=show_plot,
            hoverinfo=hoverinfo,
            n_bins=n_bins[feature],
            bins=bins[feature],
        )

        if save_directory is not None:
            save_loc = fig.save_fig(save_directory)
            print(f"[{i+1}/{len(features)}] Saving univariate anaylsis for {feature}")
            save_locs.append(save_loc)

        figs[feature] = fig

    return figs, save_locs


def build_powerpoint(
    fig_locs: list,
    feature_names: list,
    save_path: pathlib.Path(),
):
    pres = Presentation()

    for i, fig_loc in enumerate(fig_locs):
        s_register = pres.slide_layouts[7]
        s = pres.slides.add_slide(s_register)

        pic = s.shapes.add_picture(
            str(pathlib.Path(fig_loc).resolve()),
            Inches(0.5),
            Inches(1.75),
            width=Inches(7),
            height=Inches(5),
        )
        pic.left = int((pres.slide_width - pic.width) / 2)

        title = s.shapes.title
        title.text = f"{i} Univariate analysis for {(feature_names)[i]}"
        title_para = s.shapes.title.text_frame.paragraphs[0]
        title_para.font.size = Pt(24)

    pres.save(save_path)



from plotly.subplots import make_subplots
from dataclasses import dataclass, field
from plottah.plots import PlotProtocol
from typing import Protocol
import pathlib

import pandas as pd
import logging


# going after a factory design pattern here
class PlotHandlerProtocol(Protocol):
    """
    A Protocol that lies out how handler should put together various PlotProtocol objects
    """

    def build_subplot(self):
        """
        function that - given a PlotProtocol object - add the plot data to the figure
        """

    def build(self):
        """
        function that - given a PlotProtocol object - add the plot data to the figure
        """

    def show(self):
        """
        show plot
        """

        ...

    def save_fig(self):
        """
        save plot
        """

        ...

    def get_fig(self):
        """
        Return figure object
        """
        ...


@dataclass
class PlotHandler:
    """
    class combining multiple PlotProtocol objects into a single figure for a single variable

    Args:
        feature_col: feature to analyze
        target_col: target column in dataframe
        specs (Optional): figure layout
        nan_count_specs (Optional):
        plot_title (Optional): p

    """

    feature_col: str
    target_col: str
    specs: list = field(
        default_factory=lambda: [[{}, {}], [{"colspan": 2, "secondary_y": True}, None]]
    )
    plot_title: str = field(default_factory=lambda: None)

    def __post_init__(self):
        """
        Plotly does not make it easy to find the correct axes-references for each subplot.

        If you have a standard 2x2 plot with 4 plots with a single y-axes, the xref and yref for each plot will be:
        [(x1, y1), (x2, y2)],
        [(x3, y3), (x4, y4)]

        If you have a  2x2 plot, where the bottom plot spans the whole row and has a secondary y_axis, the xref and yref
        [(x1, y1), (x2, y2)],
        [(x3, y3 & y4), (None, None)]

        If you have a  2x2 plot, where the top plot spans the whole row and has a secondary y_axis, the xref and yref
        [(x1, y1 & y2), (None, None)],
        [(x2, y3), (x3, y4)]

        This last option means that we cannot simply use the row and column number to map to xref - the 2,1 element does not always map to x3.

        This post init method uses the specs to obtain a map from the position of the subplot to what x and y ref it should use.

        """
        # check if default spec was used
        self.default_spec = self.specs == [
            [{}, {}],
            [{"colspan": 2, "secondary_y": True}, None],
        ]

        # make these in init later
        self.nrows = len(self.specs)
        self.ncols = len(self.specs[0])

        # use default title if not provided
        self.plot_title = (
            f"{self.feature_col}: Roc Curve | Densities | Event Rates"
            if self.plot_title is None
            else self.plot_title
        )

        counter = 1
        self.xrefs = [
            [f"x?" for j, el in enumerate(row)] for i, row in enumerate(self.specs)
        ]

        self.yrefs = [
            [f"y{i * self.ncols + j + 1}" for j, el in enumerate(row)]
            for i, row in enumerate(self.specs)
        ]

        self.yaxes = [
            [f"yaxis{i * self.ncols + j + 1}" for j, el in enumerate(row)]
            for i, row in enumerate(self.specs)
        ]

        self.xaxes = [
            [f"xaxis{i * self.ncols + j + 1}" for j, el in enumerate(row)]
            for i, row in enumerate(self.specs)
        ]

        rows = range(self.nrows)
        cols = range(self.ncols)
        for i in rows:
            for j in cols:
                if self.specs[i][j] == None:
                    self.legend_xref = i
                else:
                    self.xrefs[i][j] = f"x{counter}"
                    # self.yrefs[i][j] = f'y{counter}'
                    counter += 1

        # make figure - access Plotly internals here
        self.fig = make_subplots(
            rows=self.nrows,
            cols=self.ncols,
            vertical_spacing=0.15,
            horizontal_spacing=0.1,
            specs=self.specs,
        )

        # update layout of figure
        self.fig.update_layout(
            showlegend=True,
            title=dict(
                text=self.plot_title,
                font=dict(size=20),
            ),
            margin=dict(t=40),
            title_x=0.5,
            width=500 * self.ncols,
            height=400 * self.nrows,
            legend=dict(
                # the below is through experimentation
                x=1.30,
                y=0.36 + (1 - self.legend_xref) * 0.57,
                xanchor="right",
                yanchor="bottom",
            ),
        )

    def build_subplot(self, subplot: PlotProtocol, row: int, col: int):
        """
        function that - given a PlotProtocol object - add the plot data to the figure

        Args:
            self: class instance containing the figure to add plots to
            subplot (PlotProtocol): plot object containing data to add to figure
            row, col (int): where on figure to add subplot

        Returns:
            None: but updates self.fig

        """
        # add traces
        logging.info(f"Building traces for {self.feature_col}")
        for trace_dict in subplot.get_traces():
            self.fig.add_trace(
                trace_dict["trace"],
                secondary_y=trace_dict["secondary_y"],
                row=row,
                col=col,
            )

        # add annotations
        logging.info(f"Building annotations for {self.feature_col}")
        for annotation in subplot.get_annotations(
            self.xrefs[row - 1][col - 1], self.yrefs[row - 1][col - 1]
        ):
            self.fig.add_annotation(**annotation)

        # update axes layout if specifed
        logging.info(f"Building x axes for {self.feature_col}")
        if subplot.get_x_axes_layout(row, col) is not None:
            self.fig.update_xaxes(**subplot.get_x_axes_layout(row, col))

        logging.info(f"Building y axes for {self.feature_col}")
        if subplot.get_y_axes_layout(row, col) is not None:
            self.fig.update_yaxes(**subplot.get_y_axes_layout(row, col))

        # update secondary y_axis if applicable

        if subplot.get_secondary_y_axis_title() is not None:
            logging.info(f"Building secondary y-axes title for {self.feature_col}")
            # name of axis contained in self.yaxes on index (row - 1, col - 1) is the primary axes, plotly will store the secondary yaxis one further; i.e. stored at index (row - 1 , col)
            secondary_yaxis = self.yaxes[row - 1][col]
            logging.info(f"Got secondary y-axes title for {self.feature_col}")
            self.fig.layout[
                secondary_yaxis
            ].title.text = subplot.get_secondary_y_axis_title()
            logging.info(f"Got layout y-axes title for {self.feature_col}")

    def build(
        self,
        df: pd.DataFrame,
        feature_col: str,
        target: str,
        topleft: PlotProtocol,
        topright: PlotProtocol,
        bottom: PlotProtocol,
        show_fig: bool = True,
    ):
        """
        function that - given a PlotProtocol object - add the plot data to the figure

        Args:
            df (pd.DataFrame): dataframe containing columns to analyze,
            feature_col (str): feature column,
            target (str): target column:
            topleft (PlotProtocol): Plot object to store in the top-left position of the plot
            topright (PlotProtocol): Plot object to store in the top-right position of the plot
            bottom (PlotProtocol): Plot object to store in the bottom row of the plot
            show_fig (bool, default = True): Whether or not to show the plot

        Returns:
            None: but updates self.fig

        """
        if not self.default_spec:
            raise ValueError(
                "build function only works for default specs, use build_subplot with each individual subplot for your specs"
            )

        # do the math for each subplot
        topleft.do_math(df, feature_col, target)
        topright.do_math(df, feature_col, target)
        bottom.do_math(df, feature_col, target)

        # build each of the subplots
        logging.info("Building topleft subplot")
        self.build_subplot(topleft, 1, 1)
        logging.info("Building topright subplot")
        self.build_subplot(topright, 1, 2)
        logging.info("Building bottom subplot")
        self.build_subplot(bottom, 2, 1)
        logging.info("Finished building subplots")

        if show_fig:
            self.show()

    def show(self):
        """
        show plot
        """

        self.fig.show()

    def save_fig(self, directory: pathlib.Path):
        """
        save plot
        """

        # make directory if not exists
        if not directory.exists():
            raise ValueError(f"directory: {directory} does not exist")

        # save the image in the directory
        self.fig.write_image(directory / f"univariate_plot_{self.feature_col}.jpeg")

        return str(directory / f"univariate_plot_{self.feature_col}.jpeg")

    def get_fig(self):
        """
        Return figure object
        """
        return self.fig



import pathlib
import pandas as pd
from typing import Dict

from plottah.plots import DistPlot, RocCurvePlot, BinEventRatePlot
from plottah.plot_handler import PlotHandler
from plottah.colors import PlotColors

#### Responsibility: put together specific Plots and Handlers into graphs


def build_standard_numerical_univariate_plot(
    df,
    feature_col: str,
    target: str,
    feature_type: str = "float",
    colors: PlotColors = PlotColors(),
    show_plot: bool = False,
    hoverinfo="all",
    n_bins: int = 10,
    bins: list = None,
    specs: list = None,
) -> PlotHandler:
    """
    buils standard univariate plot from days 'ye

    Returns
    """
    # build all 3 subplots for general plot

    ### CREATION -> should not be here? But this is like main, here we put things together ###
    roc_plot = RocCurvePlot(hoverinfo=hoverinfo, colors=colors)
    dist_plot = DistPlot(hoverinfo=hoverinfo, colors=colors)
    event_plot = BinEventRatePlot(
        hoverinfo=hoverinfo,
        colors=colors,
        n_bins=n_bins,
        bins=bins,
        feature_type=feature_type,
    )

    # need different specs, user can override but not required
    specs = (
        [[{}, {}], [{"colspan": 2, "secondary_y": True}, None]]
        if specs is None
        else specs
    )

    plot = PlotHandler(
        feature_col=feature_col, target_col=target, specs=specs, plot_title=None
    )

    ### USE ###
    # set show fig to false, show explicitly below
    plot.build(
        df=df,
        feature_col=feature_col,
        target=target,
        topleft=roc_plot,
        topright=dist_plot,
        bottom=event_plot,
        show_fig=False,
    )

    return plot


def build_standard_categorical_univariate_plot(
    df,
    feature_col: str,
    target: str,
    feature_type: str = "categorical",
    colors: PlotColors = PlotColors(),
    show_plot: bool = False,
    hoverinfo="all",
    n_bins: int = 10,
    bins: list = None,
    specs: list = None,
) -> PlotHandler:
    """
    buils standard univariate plot from days 'ye

    Returns
    """

    # create plot and do math
    event_plot = BinEventRatePlot(
        hoverinfo=hoverinfo,
        colors=colors,
        n_bins=n_bins,
        bins=bins,
        feature_type=feature_type,
    )
    event_plot.do_math(df, feature_col, target)

    # need different specs, user can override but not required
    specs = [[{"colspan": 2, "secondary_y": True}, None]] if specs is None else specs

    # set up handler and build only subplot
    plot = PlotHandler(
        feature_col=feature_col,
        target_col=target,
        specs=specs,
        plot_title=f"{feature_col}: Event Rates",
    )
    plot.build_subplot(event_plot, 1, 1)

    return plot


# Map
PLOT_BUILDERS_DICT = {
    "categorical": build_standard_categorical_univariate_plot,
    "numerical": build_standard_numerical_univariate_plot,
}



import plotly.graph_objects as go
import numpy as np
from dataclasses import dataclass, field
from sklearn import metrics

from plottah.plots.PlotProtocol import PlotProtocol
from plottah.utils import remove_or_impute_nan_infs
from plottah.colors import PlotColors

import logging 

@dataclass
class RocCurvePlot(PlotProtocol):
    # set the colorway
    colors: PlotColors = field(default_factory=lambda: PlotColors())

    # set hover setting
    hoverinfo: str = field(default_factory=lambda: "skip")

    def do_math(self, df, feature_col, target_col, fillna: bool = False):
        """
        does the required math to generate the traces, annotations and axes for the roc-curve plot

        1. imputes/removes missing values
        2. calculated fpr, tpr and AUC
        """

        logging.info("Started math for RocCurvePlot")

        # 1. impute/remove missing values
        self.df_imputed = remove_or_impute_nan_infs(df.copy(), feature_col, target_col)

        # 2. calculate fpr, tpr and AUC
        self.fpr, self.tpr, _ = metrics.roc_curve(
            self.df_imputed[target_col], -self.df_imputed[feature_col], pos_label=1
        )
        self.auc = metrics.auc(self.fpr, self.tpr)

        if self.auc < 0.5:
            self.fpr, self.tpr, _ = metrics.roc_curve(
                self.df_imputed[target_col], self.df_imputed[feature_col], pos_label=1
            )
            self.auc = metrics.auc(self.fpr, self.tpr)

    def get_traces(self):
        return [
            # plot the roc-curve
            {
                "trace": go.Scatter(
                    x=self.fpr,
                    y=self.tpr,
                    mode="lines",
                    line=dict(
                        color=self.colors.get_rgba(),
                        width=1.5,
                    ),
                    hoverinfo=self.hoverinfo,
                    showlegend=False,
                ),
                # share y
                "secondary_y": False,
            },
            # plot the baseline
            {
                "trace": go.Scatter(
                    x=np.linspace(0, 1, 10),
                    y=np.linspace(0, 1, 10),
                    mode="lines",
                    line=dict(
                        color=self.colors.get_grey_rgba(),
                        dash="dash",
                        width=0.5,
                    ),
                    hoverinfo=self.hoverinfo,
                    showlegend=False,
                ),
                # share y
                "secondary_y": False,
            },
        ]

    def get_x_axes_layout(self, row, col):
        return dict(
            title_text="Cumulated negatives",
            title_font={"size": 12},
            range=[0, 1],
            row=row,
            col=col,
            title_standoff=5,  # decrease space between title and plot
        )

    def get_y_axes_layout(self, row, col):
        return dict(
            title_text="Cumulated positives",
            title_font={"size": 12},
            range=[0, 1.05],
            row=row,
            col=col,
            title_standoff=5,  # decrease space between title and plot
        )

    def get_annotations(self, xref, yref):
        return [
            dict(
                x=0.65,
                y=0.1,
                xref=xref,
                yref=yref,
                text=f"area = {self.auc:.3f}",
                showarrow=False,
                bordercolor="rgba(255,255,255,1)",
                borderwidth=2,
                borderpad=4,
                bgcolor="rgba(255,255,255,1)",
                opacity=0.8,
            )
        ]



from dataclasses import dataclass


@dataclass
class PlotColors:
    """
    container for colors to use
    """

    primary_color: str = "40, 186, 116"
    secondary_color: str = "41, 94, 126"
    tertiary_color: str = "153, 204, 235"
    grey_tint_color: str = "110, 111, 115"

    def __post_init__(self):
        self.colors = {
            "primary_color": self.primary_color,
            "secondary_color": self.secondary_color,
            "tertiary_color": self.tertiary_color,
            "grey_tint_color": self.grey_tint_color,
        }

    def get_rgba(self, color: str = "primary_color", opacity: float = 1):
        if color not in self.colors.keys():
            raise ValueError(
                f"{color} is not one of the colors, choose from: {list(self.colors.keys())}"
            )

        return "rgba(" + self.colors[color] + f", {opacity})"

    def get_grey_rgba(self, opacity: float = 1):
        return "rgba(" + self.colors["grey_tint_color"] + f", {opacity})"


SOMEBANK_COLORS = PlotColors(
    primary_color="231, 30, 87",
    secondary_color="153, 204, 235",
    tertiary_color="254, 189, 64",
    grey_tint_color="110, 111, 115",
)

BCG_COLORS = PlotColors(
    primary_color="40, 186, 116",
    secondary_color="41, 94, 126",
    tertiary_color="153, 204, 235",
    grey_tint_color="110, 111, 115",
)



import plotly.graph_objects as go
import plotly.figure_factory as ff
from dataclasses import dataclass, field

from .PlotProtocol import PlotProtocol
from plottah.utils import remove_or_impute_nan_infs
from plottah.colors import PlotColors

import logging


@dataclass
class DistPlot(PlotProtocol):
    # set the colorway
    colors: PlotColors = field(default_factory=lambda: PlotColors())

    # set hover setting
    hoverinfo: str = field(default_factory=lambda: "skip")

    def do_math(self, df, feature_col, target_col, fillna: bool = False):
        """
        does the required math to generate the traces, annotations and axes for the roc-curve plot

        1. imputes/removes missing values
        2. extract traces from the distplot function from plotly
        3. get the max density and feature value after imputing
        """

        logging.info("Started math for DistPlot")

        # 1. impute/remove missing values
        self.df_imputed = remove_or_impute_nan_infs(df.copy(), feature_col, target_col)

        # 2. extract traces from the distplot function from plotly
        self.hist_data = [
            self.df_imputed.loc[(self.df_imputed[target_col] == 0), feature_col].values,
            self.df_imputed.loc[(self.df_imputed[target_col] == 1), feature_col].values,
        ]
        self.group_labels = ["0", "1"]
        self.distplot = ff.create_distplot(self.hist_data, self.group_labels)

        # 3. get the max density and feature value after imputing
        self.max_density = max(
            self.distplot["data"][2].y.max(), self.distplot["data"][3].y.max()
        )
        self.max_val_adj = self.df_imputed[feature_col].max()

    def get_traces(self):
        return [
            # plot the first distribution:
            {
                "trace": go.Scatter(
                    self.distplot["data"][2],
                    line=dict(color=self.colors.get_rgba(), width=0.5),
                    fill="tonexty",
                    fillcolor=self.colors.get_rgba(opacity=0.2),
                    hoverinfo=self.hoverinfo,
                ),
                # share y
                "secondary_y": False,
            },
            # plot the second distribution
            {
                "trace": go.Scatter(
                    self.distplot["data"][3],
                    line=dict(color=self.colors.get_rgba("secondary_color"), width=0.5),
                    fill="tozeroy",
                    fillcolor=self.colors.get_rgba("secondary_color", opacity=0.2),
                    hoverinfo=self.hoverinfo,
                ),
                # share y
                "secondary_y": False,
            },
        ]

    def get_x_axes_layout(self, row, col):
        return None

    def get_y_axes_layout(self, row, col):
        return dict(
            title_text="Density",
            title_font={"size": 12},
            row=row,
            col=col,
            title_standoff=5,  # decrease space between title and plot
        )

    def get_annotations(self, xref, yref):
        return [
            dict(
                x=0.9 * self.max_val_adj,
                y=1 * self.max_density,
                xref=xref,
                yref=yref,
                text=f"Class: 0",
                font=dict(color=self.colors.get_rgba()),
                showarrow=False,
                bordercolor="rgba(255,255,255,1)",
                borderwidth=2,
                borderpad=4,
                bgcolor="rgba(255,255,255,1)",
                opacity=0.8,
            ),
            dict(
                x=0.9 * self.max_val_adj,
                y=0.9 * self.max_density,
                xref=xref,
                yref=yref,
                text=f"Class: 1",
                font=dict(color=self.colors.get_rgba("secondary_color")),
                showarrow=False,
                bordercolor="rgba(255,255,255,1)",
                borderwidth=2,
                borderpad=4,
                bgcolor="rgba(255,255,255,1)",
                opacity=0.8,
            ),
        ]



from typing import Dict, List
from dataclasses import dataclass, field
import logging

import plotly.graph_objects as go
import numpy as np
import pandas as pd

from plottah.plots.PlotProtocol import PlotProtocol
from plottah.colors import PlotColors
from plottah.utils import get_bins, get_min_max_adj, get_labels_from_bins

MIN_N_UNIQUE = 25


@dataclass
class CategoricalBinner:
    """
    seperated out binning to increase cohesion -> only responsible for
        1. Adding bins to dataframe given as input
        2. Returning labels based on bins

    """

    _labels: list = field(default_factory=lambda: None)

    def get_labels(self):
        return self._labels.copy()

    def add_bins(self, df: pd.DataFrame, feature_col: str) -> pd.DataFrame:
        # Extract unique values
        unique_vals = df[feature_col].sort_values().unique()

        # Convert sorted elements to int
        # mapping = pd.factorize(unique_vals, na_sentinel=len(unique_vals))
        mapping = pd.factorize(unique_vals, na_sentinel=None)

        # Create mapping
        mapping_values = list(mapping[0])
        mapping_keys = list(mapping[1])

        # Create mapping dictionary
        mapping_dict = dict(zip(mapping_keys, mapping_values))

        # Apply mapping to column
        df = df.assign(bins=df[feature_col].map(mapping_dict))

        # Set labels for plotting
        self._labels = [str(key) for key in mapping_keys]
        if df["bins"].isna().sum() > 0:
            self._labels.append("NA")

        return df, self._labels


@dataclass
class StandardBinner:
    """
    seperated out binning to increase cohesion -> only responsible for
        1. Adding bins to dataframe given as input
        2. Returning labels based on bins

    """

    _labels: list = field(default_factory=lambda: None)

    def get_labels(self):
        return self._labels.copy()

    def add_bins(
        self,
        df: pd.DataFrame,
        feature_col: str,
        n_bins: int,
        bins=None,
        method="quantile",
    ) -> pd.DataFrame:
        # Get unadjusted min and max
        min_val, max_val = get_min_max_adj(df, feature_col)
        min_val_adj, max_val_adj = get_min_max_adj(df, feature_col)

        # boolean checking if bins were provided
        provided_bins = bins is not None

        ## BINNING
        bins = (
            bins
            if provided_bins
            else get_bins(
                df,
                feature_col,
                min_val_adj,
                max_val_adj,
                n_bins=n_bins,
                method=method,
            )
        )

        # convert type to list
        bins = list(bins)

        # set first and last value back to min and max before imputing
        if (bins[0] != min_val) & provided_bins:
            logging.warning(
                f"For feature {feature_col} you provided custom bins, the smallest bin provided is {bins[0]} but minimum is {min_val}, so replacing {bins[0]} by {min_val}"
            )
            bins[0] = min_val

        if (bins[n_bins - 1] != max_val) & provided_bins:
            logging.warning(
                f"For feature {feature_col} you provided custom bins, the largest bin is {bins[n_bins - 1] } but max is {max_val}, so replacing {bins[n_bins - 1]} by {max_val}"
            )
            bins[n_bins - 1] = max_val

        # error if not all bins are smaller than the last bin after replacing last bin by max
        if (not all(i < bins[n_bins - 1] for i in bins[:-1])) & provided_bins:
            logging.warning(
                f"For feature {feature_col} you provided custom bins, the largest bin is {bins[n_bins - 1]} but other bins provided by you are larger, and since bins must increase monotonically, this will lead to errors. Please revise the bins.Please revise the bins. Currently (after replacing smallest and largest bin by min/max): {bins}"
            )

        # error if not all bins are larger than the first bin after replacing first bin by min
        if (not all(i > bins[0] for i in bins[1:])) & provided_bins:
            logging.warning(
                f"For feature {feature_col} you provided custom bins, the smallest bin is {bins[0]} but other bins provided by you are smaller, and since bins must increase monotonically, this will lead to errors. Please revise the bins. Currently (after replacing smallest and largest bin by min/max): {bins}"
            )
        logging.info(f"using bins: {bins}")

        # update number of bins

        # Create bins (return None if binning is not successfull)
        try:
            # create new column indicating what bin record belongs to
            df = df.assign(
                bins=pd.cut(
                    x=df.loc[:, feature_col],
                    bins=bins,
                    include_lowest=True,
                    right=True,
                    labels=False,
                )
            )
        except Exception as e:
            raise ValueError(f"{self.feature_col} cannot be binned using bins: {bins}")

        # Create plot labels: [(4, 6), (6, 10), ...]
        self._labels = get_labels_from_bins(bins)
        logging.info(f"using labels: {self._labels}")

        ## CLIPPING

        # Clip values according to min/max values
        df[feature_col].clip(lower=min_val, upper=max_val, inplace=True)

        # Ensure clipping values does not remove all but a single value
        if df[feature_col].nunique() < 2:
            raise ValueError(
                "{self.feature_col} contains less than 2 features after clipping outliers!!"
            )

        ## NA's

        # Handle NAs
        if df["bins"].isna().sum() > 0:
            # replace the NA bin w/ n_bins - 1
            df.loc[:, "bins"] = df.loc[:, "bins"].where(
                ~df.loc[:, "bins"].isna(), n_bins - 1
            )
            self._labels.append("NA")
            n_bins += 1

        # Convert bins to categories
        df.bins = df.bins.astype("category")

        # Set all categories
        df.bins = df.bins.cat.set_categories(list(range(n_bins - 1)))

        return df, self._labels


@dataclass
class BinEventRatePlot(PlotProtocol):
    # set the colorway
    colors: PlotColors = field(default_factory=lambda: PlotColors())

    # set (number of) bins
    bins: list = field(default_factory=lambda: None)
    n_bins: int = field(default_factory=lambda: 10)

    # set hover setting
    hoverinfo: str = field(default_factory=lambda: "skip")

    # set default feature type
    feature_type: str = field(default_factory=lambda: "float")

    def do_math(
        self,
        df: pd.DataFrame,
        feature_col: str,
        target_col: str,
        method: str = "quantile",
    ):
        """
        does the required math to generate the traces, annotations and axes for the roc-curve plot

        1. imputes/removes missing values
        2. extract traces from the distplot function from plotly
        3. get the max density and feature value after imputing
        """

        logging.info(f"Started math for BinEventRatePlot for {feature_col}")

        # set feature and target column names
        self.feature_col = feature_col
        self.target_col = target_col

        # make fresh copy of df
        self.df = df.copy()

        # Calculate global event rate
        self.event_rate = np.mean(self.df[target_col])

        # Adjust n_bins if less unique values exist
        self.n_bins = self.n_bins if self.bins is None else len(self.bins)
        n_unique_feat_vals = df[feature_col].nunique()
        if n_unique_feat_vals < self.n_bins:
            logging.warning(
                f"{self.feature_col} only has {n_unique_feat_vals} distinct values, decreasing n_bins from {self.n_bins} to {n_unique_feat_vals} "
            )
        self.n_bins = np.minimum(n_unique_feat_vals, self.n_bins)

        # add bins column using strategy depending on feature type
        if self.feature_type == "categorical":
            logging.info("using categorical binner")
            if self.df[self.feature_col].nunique() > self.n_bins:
                raise ValueError(
                    f"Too many unique values for feature: {self.feature_col} ({self.df[self.feature_col].nunique()}) to only use {self.n_bins} bins. Increase n_bins to at least {self.df[self.feature_col].nunique()}!"
                )
            binner = CategoricalBinner()
            self.df, self.labels = binner.add_bins(self.df, self.feature_col)
        else:
            logging.info("Using standard binner")
            if self.df[self.feature_col].nunique() < MIN_N_UNIQUE:
                logging.warning(
                    f"{self.feature_col} only has {self.df[self.feature_col].nunique()} distinct values, consider switching feature type for {self.feature_col} to categorical (currenly {self.feature_type})"
                )
            binner = StandardBinner()
            self.df, self.labels = binner.add_bins(
                self.df, self.feature_col, self.n_bins, self.bins, method=method
            )

        # Group into bins and calculate required metrics - observed = False doesn't impact anything but needed for warning
        self.df_binned = self.df.groupby("bins", observed=False).agg(
            {feature_col: [len], target_col: ["mean"]}
        )

        # Rename columns
        level_one = self.df_binned.columns.get_level_values(0).astype(str)
        level_two = self.df_binned.columns.get_level_values(1).astype(str)
        column_separator = ["_" if x != "" else "" for x in level_two]
        self.df_binned.columns = level_one + column_separator + level_two

        # Set NA counts to zero
        self.df_binned[f"{feature_col}_len"] = self.df_binned[
            f"{feature_col}_len"
        ].fillna(0)

        # make fractions
        self.df_binned[f"{feature_col}_len"] = (
            self.df_binned[f"{feature_col}_len"]
            / self.df_binned[f"{feature_col}_len"].sum()
        )

    def get_traces(self) -> List[Dict]:
        return [
            # plot bar chart
            {
                "trace": go.Bar(
                    x=self.labels,
                    y=self.df_binned[f"{self.feature_col}_len"],
                    marker_color=self.colors.get_rgba("secondary_color", opacity=0.1),
                    marker_line_color=self.colors.get_rgba("secondary_color"),
                    marker_line_width=1.5,
                    opacity=0.6,
                    showlegend=False,
                    hoverinfo=self.hoverinfo,
                ),
                # share y
                "secondary_y": False,
            },
            # plot binned event rate baseline
            {
                "trace": go.Scatter(
                    x=self.labels,
                    y=[self.event_rate] * len(self.labels),
                    mode="lines",
                    line=dict(
                        color=self.colors.get_grey_rgba(),
                        dash="dash",
                        width=1,
                    ),
                    hoverinfo=self.hoverinfo,
                    name=f"General Event Rate: ({'{:.1%}'.format(self.event_rate)})",
                ),
                # share y
                "secondary_y": True,
            },
            # plot binned event rate
            {
                "trace": go.Scatter(
                    x=self.labels,
                    y=self.df_binned[f"{self.target_col}_mean"],
                    mode="lines+markers",
                    line=dict(
                        color=self.colors.get_rgba(),
                        width=1,
                    ),
                    hoverinfo=self.hoverinfo,
                    name="Event Rate",
                ),
                # share y
                "secondary_y": True,
            },
        ]

    def get_x_axes_layout(self, row, col):
        return dict(
            title_text=f"{self.feature_col}",
            title_font={"size": 12},
            row=row,
            col=col,
            title_standoff=5,  # decrease space between title and plot
            tickangle=22.5,
        )

    def get_y_axes_layout(self, row, col):
        return dict(
            title_text="Fraction of Observations",
            title_font={"size": 12},
            # range=[0, 1.2 * self.df_binned[f"{self.feature_col}_len"].max()],
            row=row,
            col=col,
            title_standoff=5,  # decrease space between title and plot
        )

    def get_secondary_y_axis_title(self):
        return "Event Rate"

    def get_annotations(self, xref, yref):
        return []



from typing import Protocol, List
from plotly.subplots import make_subplots

import plotly.graph_objects as go


class PlotProtocol(Protocol):
    """
    Defines an interface for all the plots that can be combined in the OtaPlotter Object

    (Cohesion) Plot classes are responsible for:

        (do_math: RUN BEFORE PLOTTING)
        - Accepting a dataframe, a feature and target column and doing all the math required to generate a plot.
          If a lot of code is required here, advice is to seperate out into _function_name functions

        (get_traces)
        -  Returning a list of dictionaries, with one entry per trace to plot where the keys are
            1. trace: the trace to add
            2. secondary_y: whether the trace is on the secondary y-axes

        (get_x_axes_layout)
        -  Return a dictionary with x-axes settings

        (get_y_axes_layout)
        -  Return a dictionary with y-axes settings

        (get_annotation)
        -  Return a list of dictionaries containing settings for annotations

        (get_secondary_y_axis_title)
        - Set the secondary y-axis title (only required if secondary y-axis exists)

        (show_plot)
        - Function that creates figure and adds all elemets of Plot to figure, mainly for development/debugging
          of individual classes

    """

    def do_math(self, df, feature_col, target_col, fillna: bool = False):
        """
        does the required math to generate the traces, annotations and axes for the roc-curve plot
        """
        ...

    def get_traces(self) -> List[dict]:
        ...

    def get_x_axes_layout(self, row: int, col: int) -> dict:
        ...

    def get_y_axes_layout(self, row: int, col: int) -> dict:
        ...

    def get_annotations(self, ref: str) -> List[dict]:
        ...

    def get_secondary_y_axis_title(self):
        return None

    def show_plot(self, show_figure: bool = True) -> go.Figure:
        # create figure with single graph
        figure = make_subplots(rows=1, cols=1, specs=[[{"secondary_y": True}]])
        row, col = 1, 1

        # add traces
        for trace_dict in self.get_traces():
            figure.add_trace(
                trace_dict["trace"],
                secondary_y=trace_dict["secondary_y"],
            )

        # add annotations
        for annotation in self.get_annotations("x1", "y1"):
            figure.add_annotation(**annotation)

        # update axes layout if specifed
        if self.get_x_axes_layout(row, col) is not None:
            figure.update_xaxes(**self.get_x_axes_layout(row, col))

        if self.get_y_axes_layout(row, col) is not None:
            figure.update_yaxes(**self.get_y_axes_layout(row, col))

        # update secondary y_axis if applicable
        if self.get_secondary_y_axis_title() is not None:
            print(self.get_secondary_y_axis_title())
            figure.layout["yaxis2"].title.text = self.get_secondary_y_axis_title()

        if show_figure:
            figure.show()

        return figure



import numpy as np
import pandas as pd

from typing import Tuple
import logging


def get_n_quantile_bins(
    df: pd.DataFrame,
    feature_col: str,
    min_val_adj: float,
    max_val_adj: float,
    n_bins: int,
) -> np.ndarray:
    """ensures we use the appropriate number of quantiles to get nbin bins

    Example:
        For a heavily skewed distribution, when using 10 quantiles, we might only get 2 values
        for example with [0,0,0,0,0,0,0,0,0,0,0,0,0,0,1], all but the last quantile will be 0, and the only bin
        will be (0,1).

    Args:
        df: pd.DataFrame,
        feature_col: str,
        min_val_adj: float,
        max_val_adj: float,
        n_bins: int

    Returns:
        Bins
    """

    # start count at number of bins
    i = n_bins

    bins = np.unique(
        df[feature_col].clip(min_val_adj, max_val_adj).quantile(np.linspace(0, 1, i))
    )

    # if after creating n_bin quantiles, we don't have n_bin bins, increase number of quantiles
    while (len(bins)) < n_bins:
        bins = np.unique(
            df[feature_col]
            .clip(min_val_adj, max_val_adj)
            .quantile(np.linspace(0, 1, i))
        )
        i += 1

    return bins


def remove_or_impute_nan_infs(
    df: pd.DataFrame, feature_col: str, target_col: str, fillna: bool = False
):
    """
    remove nan, inf and -inf elements or impute w/ median
    """
    if fillna == True:
        df.loc[df[feature_col].isin([-np.inf, np.inf, np.nan]), feature_col] = df.loc[
            ~df[feature_col].isin([-np.inf, np.inf, np.nan]), feature_col
        ].median()
    else:
        df = df.loc[~df[feature_col].isin([-np.inf, np.inf, np.nan])]

    return df


def get_bins(
    df: pd.DataFrame,
    feature_col: str,
    min_val_adj: float,
    max_val_adj: float,
    n_bins: int = 10,
    method: str = "quantile",
    bins: list = None,
) -> np.ndarray:
    """
    creates bins using various methods
    """

    # check: if method not manual then must be quantile or linear
    if method not in ["quantile", "linear"]:
        raise ValueError(
            f"{method} not one of the allowed methods, must be in [quantile, linear]"
        )

    # Create equidistant grid
    if method == "quantile":
        # use this function to ensure we get nbin bins even when using quantiles
        bins = get_n_quantile_bins(df, feature_col, min_val_adj, max_val_adj, n_bins)

    else:
        # get bins using linspace
        bins = np.linspace(min_val_adj, max_val_adj, n_bins)

    return bins


def get_min_max(df: pd.DataFrame, feature_col: str) -> Tuple[float, float]:
    """
    get min and max while ignoring nan values

    see: https://stackoverflow.com/questions/62865647/why-is-max-and-min-of-numpy-array-nan
    """

    # Get unadjusted min and max
    min_val = np.nanmin(df[feature_col])
    max_val = np.nanmax(df[feature_col])

    return (min_val, max_val)


def get_min_max_adj(df: pd.DataFrame, feature_col: str) -> Tuple[float, float]:
    """
    get min and max while ignoring nan values and possible +/- inf

    see: https://stackoverflow.com/questions/62865647/why-is-max-and-min-of-numpy-array-nan
    """

    # Get unadjusted min and max
    min_val = np.nanmin(df[feature_col])
    max_val = np.nanmax(df[feature_col])

    # get min & max excluding +/- inf
    idx_neg_inf = df[feature_col] == -np.inf
    idx_pos_inf = df[feature_col] == np.inf
    min_val_excl_inf = np.nanmin(df[feature_col][~idx_neg_inf])
    max_val_excl_inf = np.nanmax(df[feature_col][~idx_pos_inf])

    # take largest minimum and smallest maximum
    min_val_adj = np.maximum(min_val, min_val_excl_inf)
    max_val_adj = np.minimum(max_val, max_val_excl_inf)

    return (min_val_adj, max_val_adj)


def get_labels_from_bins(bins: list) -> list:
    """
    generate labels from bins

    example: [0, 1, 2] -> ['(0.00, 1.00]', '(1.00, 2.00]']
    """

    precision = 0
    while len(bins) > len(np.unique(np.round(bins, precision))):
        logging.info(f"precision currently: {precision}")
        precision += 1

    labels = [
        f'({"{:,.{prec}f}".format(bins[i], prec=precision)}, {"{:,.{prec}f}".format(bins[i+1], prec=precision)}]'
        for i in range(len(bins) - 1)
    ]
    return labels
